#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel detector simulation framework.

Pyxel is a detector simulation framework, that can simulate a variety of
detector effects (e.g., cosmic rays, radiation-induced CTI in CCDs, persistence
in MCT, charge diffusion, crosshatches, noises, crosstalk etc.) on a given image.
"""
import argparse
import logging
import sys
import time
import typing as t
from pathlib import Path

import dask
import numpy as np
from dask import delayed
from matplotlib import pyplot as plt
from tqdm.auto import tqdm

from pyxel import __version__ as version
from pyxel import inputs_outputs as io
from pyxel.calibration import Calibration
from pyxel.detectors import CCD, CMOS
from pyxel.dynamic import Dynamic
from pyxel.inputs_outputs import Configuration
from pyxel.parametric import Parametric
from pyxel.pipelines import DetectionPipeline, Processor
from pyxel.single import Single
from pyxel.util import download_examples

if t.TYPE_CHECKING:
    from .inputs_outputs import (
        CalibrationOutputs,
        DynamicOutputs,
        ParametricOutputs,
        Result,
        SingleOutputs,
    )


def single_mode(
    single: "Single",
    detector: t.Union["CCD", "CMOS"],
    pipeline: "DetectionPipeline",
) -> None:
    """Run a 'single' pipeline.

    Parameters
    ----------
    single
    detector
    pipeline

    Returns
    -------
    None
    """
    logging.info("Mode: Single")

    single_outputs = single.outputs  # type: SingleOutputs
    detector.set_output_dir(single_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    _ = processor.run_pipeline()

    single_outputs.save_to_file(processor)
    single_outputs.single_to_plot(processor)


def parametric_mode(
    parametric: "Parametric",
    detector: t.Union["CCD", "CMOS"],
    pipeline: "DetectionPipeline",
    with_dask: bool = False,
) -> None:
    """Run a 'parametric' pipeline.

    Parameters
    ----------
    parametric
    detector
    pipeline
    with_dask

    Returns
    -------
    None
    """
    logging.info("Mode: Parametric")

    parametric_outputs = parametric.outputs  # type: ParametricOutputs
    detector.set_output_dir(parametric_outputs.output_dir)  # TODO: Remove this

    # TODO: This should be done during initializing of object `Configuration`
    # parametric_outputs.params_func(parametric)

    processor = Processor(detector=detector, pipeline=pipeline)

    # Check if all keys from 'parametric' are valid keys for object 'pipeline'
    for param_value in parametric.enabled_steps:
        key = param_value.key  # type: str
        assert processor.has(key)

    processors_it = parametric.collect(processor)  # type: t.Iterator[Processor]

    result_list = []  # type: t.List[Result]
    output_filenames = []  # type: t.List[t.Sequence[Path]]

    # Run all pipelines
    for proc in tqdm(processors_it):  # type: Processor

        if not with_dask:
            result_proc = proc.run_pipeline()  # type: Processor
            result_val = parametric_outputs.extract_func(
                processor=result_proc
            )  # type: Result

            filenames = parametric_outputs.save_to_file(
                processor=result_proc
            )  # type: t.Sequence[Path]

        else:
            result_proc = delayed(proc.run_pipeline)()
            result_val = delayed(parametric_outputs.extract_func)(processor=result_proc)

            filenames = delayed(parametric_outputs.save_to_file)(processor=result_proc)

        result_list.append(result_val)
        output_filenames.append(filenames)  # TODO: This is not used

    if not with_dask:
        plot_array = parametric_outputs.merge_func(result_list)  # type: np.ndarray
    else:
        array = delayed(parametric_outputs.merge_func)(result_list)
        plot_array, _ = dask.compute(array, output_filenames)

    # TODO: Plot with dask ?
    if parametric_outputs.parametric_plot is not None:
        parametric_outputs.plotting_func(plot_array)


def dynamic_mode(
    dynamic: "Dynamic",
    detector: t.Union["CCD", "CMOS"],
    pipeline: "DetectionPipeline",
) -> None:
    """Run a 'dynamic' pipeline.

    Parameters
    ----------
    dynamic
    detector
    pipeline

    Returns
    -------
    None
    """

    logging.info("Mode: Dynamic")

    dynamic_outputs = dynamic.outputs  # type: DynamicOutputs

    detector.set_output_dir(dynamic_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    if isinstance(detector, CCD):
        dynamic.non_destructive_readout = False

    detector.set_dynamic(
        steps=dynamic.steps,
        time_step=dynamic.t_step,
        ndreadout=dynamic.non_destructive_readout,
    )

    # TODO: Use an iterator for that ?
    while detector.elapse_time():
        logging.info("time = %.3f s", detector.time)
        if detector.is_non_destructive_readout:
            detector.initialize(reset_all=False)
        else:
            detector.initialize(reset_all=True)
        processor.run_pipeline()
        if detector.read_out:
            dynamic_outputs.single_output(processor)


def calibration_mode(
    calibration: "Calibration",
    detector: t.Union["CCD", "CMOS"],
    pipeline: "DetectionPipeline",
) -> t.Tuple:
    """Run a 'calibration' pipeline.

    Parameters
    ----------
    calibration
    detector
    pipeline

    Returns
    -------
    None
    """
    logging.info("Mode: Calibration")

    calibration_outputs = calibration.outputs  # type: CalibrationOutputs
    detector.set_output_dir(calibration_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    ds_results, df_processors, df_all_logs = calibration.run_calibration(
        processor=processor, output_dir=calibration_outputs.output_dir
    )

    # TODO: Save the processors from 'df_processors'
    # TODO: Generate plots from 'ds_results'

    # TODO: Do something with 'df_all_logs' ?

    # TODO: create 'output' object with .calibration_outputs
    # TODO: use 'fitting.get_simulated_data' ==> np.ndarray

    # geometry = processor.detector.geometry
    # calibration.post_processing(
    #     champions=champions,
    #     output=calibration_outputs,
    #     row=geometry.row,
    #     col=geometry.col,
    # )
    filenames = calibration.post_processing(
        ds=ds_results, df_processors=df_processors, output=calibration_outputs
    )

    return ds_results, df_processors, df_all_logs, filenames


def output_directory(configuration: Configuration) -> Path:
    """Return the output directory from the configuration.

    Parameters
    ----------
    configuration

    Returns
    -------
    output_dir
    """
    if isinstance(configuration.single, Single):
        output_dir = configuration.single.outputs.output_dir
    elif isinstance(configuration.calibration, Calibration):
        output_dir = configuration.calibration.outputs.output_dir
    elif isinstance(configuration.dynamic, Dynamic):
        output_dir = configuration.dynamic.outputs.output_dir
    elif isinstance(configuration.parametric, Parametric):
        output_dir = configuration.parametric.outputs.output_dir
    else:
        raise (ValueError("Outputs not initialized."))
    return output_dir


def run(input_filename: str, random_seed: t.Optional[int] = None) -> None:
    """TBW.

    Parameters
    ----------
    input_filename
    random_seed
    """
    logging.info("Pyxel version %s", version)
    logging.info("Pipeline started.")

    start_time = time.time()
    if random_seed:
        np.random.seed(random_seed)

    configuration = io.load(
        Path(input_filename).expanduser().resolve()
    )  # type: Configuration

    output_dir = output_directory(configuration)

    io.save(input_filename=input_filename, output_dir=output_dir)

    pipeline = configuration.pipeline  # type: DetectionPipeline

    if isinstance(configuration.ccd_detector, CCD):
        detector = configuration.ccd_detector  # type: t.Union[CCD, CMOS]
    elif isinstance(configuration.cmos_detector, CMOS):
        detector = configuration.cmos_detector
    else:
        raise NotImplementedError("Detector is not defined in YAML config. file!")

    if isinstance(configuration.single, Single):
        single = configuration.single  # type: Single
        single_mode(single=single, detector=detector, pipeline=pipeline)

    elif isinstance(configuration.calibration, Calibration):

        calibration = configuration.calibration  # type: Calibration
        _ = calibration_mode(
            calibration=calibration, detector=detector, pipeline=pipeline
        )

    elif isinstance(configuration.parametric, Parametric):
        parametric = configuration.parametric  # type: Parametric
        parametric_mode(parametric=parametric, detector=detector, pipeline=pipeline)

    elif isinstance(configuration.dynamic, Dynamic):

        dynamic = configuration.dynamic  # type: Dynamic
        dynamic_mode(dynamic=dynamic, detector=detector, pipeline=pipeline)

    else:
        raise NotImplementedError("Please provide a valid simulation mode !")

    logging.info("Pipeline completed.")
    logging.info("Running time: %.3f seconds" % (time.time() - start_time))
    # Closing the logger in order to be able to move the file in the output dir
    logging.shutdown()
    io.save_log_file(output_dir)
    plt.close()


# TODO: Use library 'click' instead of 'parser' ? See issue #62
#       Add an option to display colors ? (very optional)
def main() -> None:
    """Define the argument parser and run Pyxel."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, description=__doc__
    )

    parser.add_argument(
        "-v",
        "--verbosity",
        action="count",
        default=0,
        help="Increase output verbosity (-v/-vv/-vvv)",
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version="Pyxel, version {version}".format(version=version),
    )

    parser.add_argument(
        "-c",
        "--config",
        type=str,
        help="Configuration file to load (YAML)",
    )

    parser.add_argument("-s", "--seed", type=int, help="Random seed for the framework")

    parser.add_argument(
        "--download-examples",
        nargs="?",
        const="pyxel-examples",
        help="Install examples to the specified directory, default is /pyxel-examples.",
    )

    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force flag for saving the examples.",
    )

    # parser.add_argument('-g', '--gui', default=False, type=bool, help='run Graphical User Interface')
    # parser.add_argument('-p', '--port', default=9999, type=int, help='The port to run the web server on')

    opts = parser.parse_args()

    logging_level = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG][
        min(opts.verbosity, 3)
    ]
    log_format = (
        "%(asctime)s - %(name)s - %(threadName)30s - %(funcName)30s \t %(message)s"
    )
    logging.basicConfig(
        filename="pyxel.log",
        level=logging_level,
        format=log_format,
        datefmt="%d-%m-%Y %H:%M:%S",
    )
    # If user wants the log in stdout AND in file, use the three lines below
    stream_stdout = logging.StreamHandler(sys.stdout)
    stream_stdout.setFormatter(logging.Formatter(log_format))
    logging.getLogger().addHandler(stream_stdout)

    if opts.config:
        run(input_filename=opts.config, random_seed=opts.seed)
    elif opts.download_examples:
        download_examples(foldername=opts.download_examples, force=opts.force)
    else:
        print("Define a YAML configuration file!")


if __name__ == "__main__":
    main()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""


class Optics:
    """Optical attributes of the detector and its optical system."""



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""TBW."""

import typing as t

if t.TYPE_CHECKING:
    from ..inputs_outputs import DynamicOutputs


class Dynamic:
    """TBW."""

    def __init__(
        self,
        outputs: "DynamicOutputs",
        t_step: float,
        steps: int,
        non_destructive_readout: bool = False,
    ):
        self.outputs = outputs
        self._t_step = t_step
        self._steps = steps
        self._non_destructive_readout = non_destructive_readout

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<outputs={self.outputs!r}>"

    @property
    def t_step(self):
        """TBW."""
        return self._t_step

    @property
    def steps(self):
        """TBW."""
        return self._steps

    @property
    def non_destructive_readout(self):
        """TBW."""
        return self._non_destructive_readout

    @non_destructive_readout.setter
    def non_destructive_readout(self, non_destructive_readout: bool) -> None:
        """TBW."""
        self._non_destructive_readout = non_destructive_readout



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""TBW."""
import logging
import typing as t
from pathlib import Path
from time import strftime

import attr
import h5py as h5
import numpy as np
import pandas as pd
from astropy.io import fits as fits
from matplotlib import pyplot as plt

from pyxel import __version__ as version

from .outputs import PlotArguments, PlotType, apply_run_number, update_plot

if t.TYPE_CHECKING:
    from ..detectors import Detector
    from ..parametric.parameter_values import ParameterValues
    from ..parametric.parametric import Parametric
    from ..pipelines import Processor

    class SaveToFile(t.Protocol):
        """TBW."""

        def __call__(self, data: np.ndarray, name: str) -> Path:
            """TBW."""
            ...


# TODO Specific to parametric
@attr.s(auto_attribs=True, slots=True, frozen=True)
class Result:
    """TBW."""

    result: np.ndarray  # TODO: Use a `DataFrame` ?
    plot: np.ndarray


@attr.s(auto_attribs=True, slots=True)
class ParametricPlot:
    """TBW."""

    x: str
    y: str
    plot_args: PlotArguments

    @classmethod
    def from_dict(cls, dct: dict) -> "ParametricPlot":
        """TBW."""
        return cls(x=dct["x"], y=dct["y"], plot_args=dct["plot_args"])


class ParametricOutputs:
    """TBW."""

    def __init__(
        self,
        output_folder: t.Union[str, Path],
        save_data_to_file: t.Optional[
            t.Sequence[t.Mapping[str, t.Sequence[str]]]
        ] = None,
        save_parameter_to_file: t.Optional[dict] = None,
        parametric_plot: t.Optional[ParametricPlot] = None,
    ):
        self._log = logging.getLogger(__name__)

        # self.input_file = None  # type: t.Optional[Path]

        # Parameter(s) specific for 'Parametric'
        self.parametric_plot = None  # type: t.Optional[ParametricPlot]
        if parametric_plot is not None:
            self.parametric_plot = parametric_plot

        self.parameter_keys = []  # type: t.List[str]

        self.user_plt_args = None  # type: t.Optional[PlotArguments]
        self.save_parameter_to_file = save_parameter_to_file  # type: t.Optional[dict]
        self.output_dir = (
            Path(output_folder).joinpath("run_" + strftime("%Y%m%d_%H%M%S")).resolve()
        )  # type: Path

        # TODO: Not related to a plot. Use by 'single' and 'parametric' modes.
        self.save_data_to_file = save_data_to_file or [
            {"detector.image.array": ["fits"]}
        ]  # type: t.Sequence[t.Mapping[str, t.Sequence[str]]]

        # TODO: reenable
        # if self.output_dir.exists():
        #    raise IsADirectoryError("Directory exists.")

        self.output_dir.mkdir(parents=True, exist_ok=True)

        # self.default_ax_args = {
        #     "xlabel": None,
        #     "ylabel": None,
        #     "title": None,
        #     "axis": None,
        #     "grid": False,
        #     "xscale": "linear",
        #     "yscale": "linear",
        #     "xticks": None,
        #     "yticks": None,
        #     "xlim": [None, None],
        #     "ylim": [None, None],
        #     "sci_x": False,
        #     "sci_y": False,
        # }  # type: dict
        self.default_ax_args = PlotArguments()

        self.default_plot_args = {
            "color": None,
            "marker": ".",
            "linestyle": "",
        }  # type: dict

        self.default_hist_args = {
            "bins": None,
            "range": None,
            "density": None,
            "log": False,
            "cumulative": False,
            "histtype": "step",
            "color": None,
            "facecolor": None,
        }  # type: dict
        self.default_scatter_args = {"size": None, "cbar_label": None}  # type: dict

        self.plt_args = {}  # type: dict

        # TODO: Create an object self._fig here ?
        fig, ax = plt.subplots(1, 1)

        self._fig = fig  # type: plt.Figure
        self._ax = ax  # type: plt.Axes

        plt.close(self._fig)

    def __repr__(self):
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<output_dir={self.output_dir!r}>"

    @property
    def fig(self) -> plt.Figure:
        """Get the current ``Figure``."""
        return self._fig

    def new_file(self, filename: str) -> Path:
        """TBW."""
        new_filename = self.output_dir.joinpath(filename)  # type: Path
        new_filename.touch()

        return new_filename

    # TODO: Specific to 'single_plot' ?
    def save_to_png(self, data: np.ndarray, name: str) -> Path:
        """Write array to bitmap PNG image file.

        Parameters
        ----------
        data : array
        name : str

        Returns
        -------
        Path
            TBW.
        """
        row, col = data.shape
        name = str(name).replace(".", "_")
        filename = apply_run_number(
            self.output_dir.joinpath(f"{name}_??.png")
        )  # type: Path

        dpi = 300
        self._fig.set_size_inches(min(col / dpi, 10.0), min(row / dpi, 10.0))

        # ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
        # ax.set_axis_off()
        #
        # fig.add_axes(ax)
        # plt.imshow(data, cmap="gray", extent=[0, col, 0, row])
        self._fig.savefig(filename, dpi=dpi)

        return filename

    def save_to_fits(self, data: np.ndarray, name: str) -> Path:
        """Write array to FITS file.

        Parameters
        ----------
        data
        name

        Returns
        -------
        Path
            TBW.
        """
        name = str(name).replace(".", "_")
        filename = apply_run_number(
            self.output_dir.joinpath(f"{name}_??.fits")
        ).resolve()  # type: Path

        self._log.info("Save to FITS - filename: '%s'", filename)

        hdu = fits.PrimaryHDU(data)
        hdu.header["PYXEL_V"] = (str(version), "Pyxel version")
        hdu.writeto(filename, overwrite=False, output_verify="exception")

        return filename

    def save_to_hdf(self, data: "Detector", name: str) -> Path:
        """Write detector object to HDF5 file."""
        name = str(name).replace(".", "_")
        filename = apply_run_number(self.output_dir.joinpath(f"{name}_??.h5"))
        with h5.File(filename, "w") as h5file:
            h5file.attrs["pyxel-version"] = str(version)
            if name == "detector":
                detector_grp = h5file.create_group("detector")
                for array, name in zip(
                    [
                        data.signal.array,
                        data.image.array,
                        data.photon.array,
                        data.pixel.array,
                        data.charge.frame,
                    ],
                    ["Signal", "Image", "Photon", "Pixel", "Charge"],
                ):
                    dataset = detector_grp.create_dataset(name, np.shape(array))
                    dataset[:] = array
            else:
                detector_grp = h5file.create_group("data")
                dataset = detector_grp.create_dataset(name, np.shape(data))
                dataset[:] = data
        return filename

    def save_to_txt(self, data: np.ndarray, name: str) -> Path:
        """Write data to txt file."""
        name = str(name).replace(".", "_")
        filename = apply_run_number(self.output_dir.joinpath(name + "_??.txt"))
        np.savetxt(filename, data, delimiter=" | ", fmt="%.8e")
        return filename

    def save_to_csv(self, data: pd.DataFrame, name: str) -> Path:
        """Write Pandas Dataframe or Numpy array to a CSV file."""
        name = str(name).replace(".", "_")
        filename = apply_run_number(
            self.output_dir.joinpath(name + "_??.csv")
        )  # type: Path
        try:
            data.to_csv(filename, float_format="%g")
        except AttributeError:
            np.savetxt(filename, data, delimiter=",", fmt="%.8e")
        return filename

    def save_to_npy(self, data: np.ndarray, name: str) -> Path:
        """Write Numpy array to Numpy binary npy file."""
        name = str(name).replace(".", "_")
        filename = apply_run_number(self.output_dir.joinpath(name + "_??.npy"))
        np.save(file=filename, arr=data)
        return filename

    def save_plot(self, filename: str = "figure_??") -> Path:
        """Save plot figure in PNG format, close figure and create new canvas for next plot."""
        new_filename = self.output_dir.joinpath(filename + ".png")  # type: Path
        output_filename = apply_run_number(new_filename).resolve()  # type: Path

        self._log.info("Save plot in filename '%s'", output_filename)
        self._fig.savefig(output_filename)
        # plt.close('all')
        # plt.figure()

        return output_filename

    def plot_graph(
        self, x: np.ndarray, y: np.ndarray, args: t.Optional[dict] = None
    ) -> None:
        """TBW.

        Parameters
        ----------
        x
        y
        args
        """
        ax_args0, plt_args0 = self.update_args(plot_type=PlotType.Graph, new_args=args)

        user_plt_args_dct = None  # type: t.Optional[t.Mapping]
        if isinstance(self.user_plt_args, PlotArguments):
            user_plt_args_dct = self.user_plt_args.to_dict()

        ax_args, plt_args = self.update_args(
            plot_type=PlotType.Graph,
            new_args=user_plt_args_dct,
            ax_args=ax_args0,
            plt_args=plt_args0,
        )

        self._ax.plot(
            x,
            y,
            color=plt_args["color"],
            marker=plt_args["marker"],
            linestyle=plt_args["linestyle"],
        )
        update_plot(ax_args=ax_args, ax=self._ax)
        # plt.draw()

    def plot_histogram(self, data: np.ndarray, args: t.Optional[dict] = None) -> None:
        """TBW.

        Parameters
        ----------
        data
        args
        """
        assert self.user_plt_args is not None

        user_plt_args_dct = None  # type: t.Optional[t.Mapping]
        if isinstance(self.user_plt_args, PlotArguments):
            user_plt_args_dct = self.user_plt_args.to_dict()

        ax_args0, plt_args0 = self.update_args(
            plot_type=PlotType.Histogram, new_args=args
        )
        ax_args, plt_args = self.update_args(
            plot_type=PlotType.Histogram,
            new_args=user_plt_args_dct,
            ax_args=ax_args0,
            plt_args=plt_args0,
        )
        if isinstance(data, np.ndarray):
            data = data.flatten()

        self._ax.hist(
            x=data,
            bins=plt_args["bins"],
            range=plt_args["range"],
            density=plt_args["density"],
            log=plt_args["log"],
            cumulative=plt_args["cumulative"],
            histtype=plt_args["histtype"],
            color=plt_args["color"],
            facecolor=plt_args["facecolor"],
        )

        update_plot(ax_args=ax_args, ax=self._ax)

    def plot_scatter(
        self,
        x: np.ndarray,
        y: np.ndarray,
        color: t.Optional[str] = None,
        args: t.Optional[dict] = None,
    ) -> None:
        """TBW.

        Parameters
        ----------
        x
        y
        color
        args
        """
        user_plt_args_dct = None  # type: t.Optional[t.Mapping]

        if isinstance(self.user_plt_args, PlotArguments):
            user_plt_args_dct = self.user_plt_args.to_dict()

        ax_args0, plt_args0 = self.update_args(
            plot_type=PlotType.Scatter, new_args=args
        )
        ax_args, plt_args = self.update_args(
            plot_type=PlotType.Scatter,
            new_args=user_plt_args_dct,
            ax_args=ax_args0,
            plt_args=plt_args0,
        )

        # fig = plt.gcf()
        # ax = fig.axes

        if color is not None:
            sp = self._ax.scatter(x, y, c=color, s=plt_args["size"])
            cbar = self._fig.colorbar(sp)
            cbar.set_label(plt_args["cbar_label"])

        else:
            self._ax.scatter(x, y, s=plt_args["size"])

        update_plot(ax_args=ax_args, ax=self._ax)
        # plt.draw()
        # fig.canvas.draw_idle()

    def save_to_file(self, processor: "Processor") -> t.Sequence[Path]:
        """Save outputs into file(s).

        Parameters
        ----------
        processor : Processor


        Returns
        -------
        list of ``Path``
            TBW.
        """
        save_methods = {
            "fits": self.save_to_fits,
            "hdf": self.save_to_hdf,
            "npy": self.save_to_npy,
            "txt": self.save_to_txt,
            "csv": self.save_to_csv,
            "png": self.save_to_png,
        }  # type: t.Dict[str, SaveToFile]

        filenames = []  # type: t.List[Path]

        for dct in self.save_data_to_file:  # type: t.Mapping[str, t.Sequence[str]]
            # Get first entry of `dict` 'item'
            first_item, *_ = dct.items()

            obj, format_list = first_item
            data = processor.get(obj)  # type: np.ndarray

            if format_list is not None:
                for out_format in format_list:
                    func = save_methods[out_format]  # type: SaveToFile
                    filename = func(data=data, name=obj)  # type: Path

                    filenames.append(filename)

        return filenames

    # TODO: Specific to 'parametric_plot' ?
    def params_func(self, param: "Parametric") -> None:
        """Extract all parametric keys from `param`."""
        assert self.parameter_keys is not None

        # TODO: Re-initialized 'self.parameters' ??

        for var in param.enabled_steps:  # type: ParameterValues
            if var.key not in self.parameter_keys:
                self.parameter_keys += [var.key]

        if self.save_parameter_to_file and self.save_parameter_to_file["parameter"]:
            for par in self.save_parameter_to_file["parameter"]:
                if par is not None and par not in self.parameter_keys:
                    self.parameter_keys += [par]

    # TODO: This function should be moved in `Parametric`
    # TODO: Specific to 'parametric_plot' ?
    def extract_func(self, processor: "Processor") -> Result:
        """TBW."""
        assert self.parameter_keys is not None

        # self.single_output(processor.detector)    # TODO: extract other things (optional)

        res_row = []  # type: t.List[np.ndarray]
        for key in self.parameter_keys:
            value = processor.get(key)  # type: np.ndarray
            res_row.append(value)

        # Extract all parameters keys from 'proc'
        # all_attr_getters = operator.attrgetter(self.parameter_keys)  # type: t.Callable
        # res_row = all_attr_getters(processor)  # type: t.Tuple[t.Any, ...]

        # TODO: Refactor this
        plt_row = []  # type: t.List[np.ndarray]
        if self.parametric_plot:
            for key in [self.parametric_plot.x, self.parametric_plot.y]:
                if key is not None:
                    value = processor.get(key)
                    plt_row.append(value)

        return Result(
            result=np.array(res_row, dtype=np.float),
            plot=np.array(plt_row, dtype=np.float),
        )

    # TODO: Specific to 'parametric_mode' ?
    def merge_func(self, result_list: t.Sequence[Result]) -> np.ndarray:
        """TBW."""
        assert self.parameter_keys is not None

        if self.save_parameter_to_file:
            result_array = np.array([k.result for k in result_list])
            save_methods = {
                "npy": self.save_to_npy,
                "txt": self.save_to_txt,
                "csv": self.save_to_csv,
            }  # type: t.Dict[str, SaveToFile]

            for out_format in self.save_parameter_to_file["file_format"]:
                func = save_methods[out_format]  # type: SaveToFile
                file = func(data=result_array, name="parameters")  # type: Path

                if file.suffix in (".txt", ".csv"):
                    with file.open("r+") as f:
                        content = f.read()
                        f.seek(0, 0)
                        f.write(
                            "# "
                            + "".join([pp + " // " for pp in self.parameter_keys])
                            + "\n"
                            + content
                        )
        plot_array = np.array([k.plot for k in result_list])  # type: np.ndarray
        return plot_array

    # TODO: Specific to 'parametric_plot' ?
    def plotting_func(self, plot_array: np.ndarray) -> None:
        """TBW.

        Parameters
        ----------
        plot_array : array
        """
        if not self.parametric_plot:
            raise RuntimeError

        x_key = self.parametric_plot.x  # type: str
        y_key = self.parametric_plot.y  # type: str

        self.user_plt_args = self.parametric_plot.plot_args

        x = plot_array[:, 0]  # type: np.ndarray
        y = plot_array[:, 1]  # type: np.ndarray

        par_name = x_key[
            x_key[: x_key[: x_key.rfind(".")].rfind(".")].rfind(".") + 1 :  # noqa: E203
        ]
        res_name = y_key[
            y_key[: y_key[: y_key.rfind(".")].rfind(".")].rfind(".") + 1 :  # noqa: E203
        ]

        args = {"xlabel": par_name, "ylabel": res_name}

        if isinstance(x, np.ndarray):
            x = x.flatten()
        if isinstance(y, np.ndarray):
            y = y.flatten()

        self.plot_graph(x=x, y=y, args=args)
        self.save_plot(filename="parametric_??")

    def update_args(
        self,
        plot_type: PlotType,
        new_args: t.Optional[t.Mapping] = None,
        ax_args: t.Optional[dict] = None,
        plt_args: t.Optional[dict] = None,
    ) -> t.Tuple[dict, dict]:
        """TBW."""
        if new_args is None:
            new_args = {}

        if ax_args is None:
            ax_args = self.default_ax_args.to_dict()

        if plt_args is None:
            if plot_type is PlotType.Histogram:
                plt_args = self.default_hist_args.copy()
            elif plot_type is PlotType.Graph:
                plt_args = self.default_plot_args.copy()
            elif plot_type is PlotType.Scatter:
                plt_args = self.default_scatter_args.copy()
            else:
                raise ValueError

        for key in new_args:
            if key in plt_args.keys():
                plt_args[key] = new_args[key]
            elif key in ax_args.keys():
                ax_args[key] = new_args[key]
            else:
                raise KeyError('Not valid plotting key in "plot_args": "%s"' % key)

        return ax_args, plt_args



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Stripe pattern illumination model."""

import typing as t

import numpy as np
import skimage.transform as tr

from pyxel.data_structure import Photon

if t.TYPE_CHECKING:
    from pyxel.detectors import Detector


def square_signal(n: int, lw: int, startwith: int = 0) -> list:
    """Compute a 1D periodic square signal.

    Parameters
    ----------
    n: int
        Length of the signal.
    lw:
        Width of a pulse.
    startwith
        1 to start with high level or 0 for 0.

    Returns
    -------
    out: list
        Output list.
    """
    if lw > n // 2:
        raise ValueError("Line too wide.")
    start = [startwith] * lw
    second = [1 - startwith] * lw
    pair = start + second
    num = n // len(pair)
    out = pair * num
    return out


def pattern(
    detector_shape: tuple,
    period: int = 2,
    level: float = 1,
    angle: float = 0,
    startwith: int = 0,
) -> np.ndarray:
    """Return an array of a periodic pattern.

    Parameters
    ----------
    detector_shape: tuple
        Detector shape.
    period: int
        Period of the periodic pattern in pixels.
    level: float
        Amplitude of the periodic pattern.
    angle: int
        Angle of the pattern in degrees.
    startwith: int
        1 to start with high level or 0 for 0.

    Returns
    -------
    out: np.ndarray
        Output stripe pattern.
    """

    if period < 2:
        raise ValueError("Cant set a period smaller than 2 pixels.")
    elif (period % 2) != 0:
        raise ValueError("Period should be a multiple of 2.")

    y, x = detector_shape

    n = max(y, x) * 2
    m = max(y, x) * 2

    sx = slice(n // 2 - y // 2, n // 2 + y // 2)
    sy = slice(m // 2 - x // 2, m // 2 + x // 2)

    signal = square_signal(n=n, lw=period // 2, startwith=startwith)
    signal = np.array(signal + ([1] * (n - len(signal))))[::-1]

    out = np.ones((n, m))

    for i in range(m):
        out[:, i] = signal

    out = level * out

    if angle:
        out = tr.rotate(out, angle=angle)

    out = out[sx, sy]

    return out


def stripe_pattern(
    detector: "Detector",
    period: int = 10,
    level: float = 1.0,
    angle: int = 0,
    startwith: int = 0,
) -> None:
    """Stripe pattern model.

    Parameters
    ----------
    detector: Detector
        Detector object.
    period: int
        Period of the periodic pattern in pixels.
    level: float
        Amplitude of the periodic pattern.
    angle: int
        Angle of the pattern in degrees.
    startwith: int
        1 to start with high level or 0 for 0.

    Returns
    -------
    None
    """

    photon_array = pattern(
        detector_shape=(detector.geometry.row, detector.geometry.col),
        period=period,
        level=level,
        startwith=startwith,
        angle=angle,
    )

    detector.photon = Photon(photon_array)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""

# flake8: noqa
from .charge_injection import charge_blocks
from .dark_current import dark_current_rule07
from .load_profile import charge_profile
from .photoelectrons import monte_carlo_conversion, simple_conversion, random_pos
from .tars.tars import run_tars
from .qe_map import qe_map



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel Charge class to generate electrons or holes inside detector."""

import typing as t

import numpy as np
import pandas as pd
from astropy.units import cds

from pyxel.data_structure import Particle

cds.enable()


class Charge(Particle):
    """Charged particle class defining and storing information of all electrons and holes.

    Properties stored are: charge, position, velocity, energy.
    """

    def __init__(self):
        # TODO: The following line is not really needed
        super().__init__()
        self.nextid = 0  # type: int

        self.columns = (
            "charge",
            "number",
            "init_energy",
            "energy",
            "init_pos_ver",
            "init_pos_hor",
            "init_pos_z",
            "position_ver",
            "position_hor",
            "position_z",
            "velocity_ver",
            "velocity_hor",
            "velocity_z",
        )  # type: t.Tuple[str, ...]

        self.EMPTY_FRAME = pd.DataFrame(
            columns=self.columns, dtype=np.float
        )  # type: pd.DataFrame       # todo

        self.frame = self.EMPTY_FRAME.copy()  # type: pd.DataFrame

    def add_charge(
        self,
        particle_type: str,  # TODO: Use Enum
        particles_per_cluster: t.Sequence[float],
        init_energy: t.Sequence[float],
        init_ver_position: t.Sequence[float],
        init_hor_position: t.Sequence[float],
        init_z_position: t.Sequence[float],
        init_ver_velocity: t.Sequence[float],
        init_hor_velocity: t.Sequence[float],
        init_z_velocity: t.Sequence[float],
    ) -> None:
        """Create new charge or group of charge inside the detector stored in a pandas DataFrame.

        :param particle_type:
        :param particles_per_cluster:
        :param init_energy:
        :param init_ver_position:
        :param init_hor_position:
        :param init_z_position:
        :param init_ver_velocity:
        :param init_hor_velocity:
        :param init_z_velocity:
        :return:
        """
        if not (
            len(particles_per_cluster)
            == len(init_energy)
            == len(init_ver_position)
            == len(init_ver_velocity)
        ):
            raise ValueError("List arguments have different lengths")

        elements = len(init_energy)

        # check_position(self.detector, init_ver_position, init_hor_position, init_z_position)      # TODO
        # check_energy(init_energy)             # TODO
        # Check if particle number is integer:
        # check_type(particles_per_cluster)      # TODO

        # TODO: particle_type should be a Enum class ?
        if particle_type == "e":
            charge = [-1] * elements  # * cds.e
        elif particle_type == "h":
            charge = [+1] * elements  # * cds.e
        else:
            raise ValueError("Given charged particle type can not be simulated")

        # if all(init_ver_velocity) == 0 and all(init_hor_velocity) == 0 and all(init_z_velocity) == 0:
        #     random_direction(1.0)

        # dict
        new_charge = {
            "charge": charge,
            "number": particles_per_cluster,
            "init_energy": init_energy,
            "energy": init_energy,
            "init_pos_ver": init_ver_position,
            "init_pos_hor": init_hor_position,
            "init_pos_z": init_z_position,
            "position_ver": init_ver_position,
            "position_hor": init_hor_position,
            "position_z": init_z_position,
            "velocity_ver": init_ver_velocity,
            "velocity_hor": init_hor_velocity,
            "velocity_z": init_z_velocity,
        }

        new_charge_df = pd.DataFrame(
            new_charge, index=range(self.nextid, self.nextid + elements)
        )
        self.nextid = self.nextid + elements
        self.frame = self.frame.append(new_charge_df, sort=False)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""TBW."""
import logging
import typing as t
from numbers import Number
from pathlib import Path
from time import strftime

import attr
import dask.delayed as delayed
import h5py as h5
import numpy as np
import pandas as pd
from astropy.io import fits as fits
from dask.delayed import Delayed
from matplotlib import pyplot as plt
from typing_extensions import Literal

from pyxel import __version__ as version

from .outputs import PlotArguments, PlotType, apply_run_number, update_plot

if t.TYPE_CHECKING:
    from ..calibration import ResultType
    from ..detectors import Detector
    from ..pipelines import Processor

    class SaveToFile(t.Protocol):
        """TBW."""

        def __call__(
            self, data: np.ndarray, name: str, with_auto_suffix: bool = True
        ) -> Path:
            """TBW."""
            ...


# Specific for CalibrationPlot
@attr.s(auto_attribs=True, slots=True, frozen=True)
class ChampionsPlot:
    """TBW."""

    plot_args: t.Optional[PlotArguments] = None


@attr.s(auto_attribs=True, slots=True, frozen=True)
class PopulationPlot:
    """TBW."""

    columns: t.Optional[t.Tuple[int, int]] = None  # TODO: Check this, with validator ?
    plot_args: t.Optional[PlotArguments] = None


@attr.s(auto_attribs=True, slots=True, frozen=True)
class FittingPlot:
    """TBW."""

    plot_args: t.Optional[PlotArguments] = None


@attr.s(auto_attribs=True, slots=True, frozen=True)
class CalibrationPlot:
    """TBW."""

    champions_plot: t.Optional[ChampionsPlot] = None
    population_plot: t.Optional[PopulationPlot] = None
    fitting_plot: t.Optional[FittingPlot] = None


# Define type aliases
ValidName = Literal[
    "detector.image.array", "detector.signal.array", "detector.pixel.array"
]
ValidFormat = Literal["fits", "hdf", "npy", "txt", "csv", "png"]


class CalibrationOutputs:
    """TBW."""

    def __init__(
        self,
        output_folder: t.Union[str, Path],
        save_data_to_file: t.Optional[
            t.Sequence[t.Mapping[ValidName, t.Sequence[ValidFormat]]]
        ] = None,
        save_parameter_to_file: t.Optional[dict] = None,
        calibration_plot: t.Optional[CalibrationPlot] = None,
    ):
        self._log = logging.getLogger(__name__)

        # self.input_file = None  # type: t.Optional[Path]

        # Parameter(s) specific for 'Calibration'
        self.calibration_plot = calibration_plot  # type: t.Optional[CalibrationPlot]

        self.user_plt_args = None  # type: t.Optional[PlotArguments]
        self.save_parameter_to_file = save_parameter_to_file  # type: t.Optional[dict]
        self.output_dir = (
            Path(output_folder).joinpath("run_" + strftime("%Y%m%d_%H%M%S")).resolve()
        )  # type: Path

        # if save_data_to_file is None:
        #     self.save_data_to_file = [{'detector.image.array': ['fits']}]       # type: list
        # else:
        # TODO: Not related to a plot. Use by 'single' and 'parametric' modes.
        self.save_data_to_file = save_data_to_file or [
            {"detector.image.array": ["fits"]}
        ]  # type: t.Sequence[t.Mapping[ValidName, t.Sequence[ValidFormat]]]

        # TODO: reenable
        # if self.output_dir.exists():
        #    raise IsADirectoryError("Directory exists.")

        self.output_dir.mkdir(parents=True, exist_ok=True)

        # self.default_ax_args = {
        #     "xlabel": None,
        #     "ylabel": None,
        #     "title": None,
        #     "axis": None,
        #     "grid": False,
        #     "xscale": "linear",
        #     "yscale": "linear",
        #     "xticks": None,
        #     "yticks": None,
        #     "xlim": [None, None],
        #     "ylim": [None, None],
        #     "sci_x": False,
        #     "sci_y": False,
        # }  # type: dict
        self.default_ax_args = PlotArguments()

        self.default_plot_args = {
            "color": None,
            "marker": ".",
            "linestyle": "",
        }  # type: dict

        self.default_hist_args = {
            "bins": None,
            "range": None,
            "density": None,
            "log": False,
            "cumulative": False,
            "histtype": "step",
            "color": None,
            "facecolor": None,
        }  # type: dict
        self.default_scatter_args = {"size": None, "cbar_label": None}  # type: dict

        self.plt_args = {}  # type: dict

        # TODO: Create an object self._fig here ?
        fig, ax = plt.subplots(1, 1)

        self._fig = fig  # type: plt.Figure
        self._ax = ax  # type: plt.Axes

        plt.close(self._fig)

    def __repr__(self):
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<output_dir={self.output_dir!r}>"

    @property
    def fig(self) -> plt.Figure:
        """Get the current ``Figure``."""
        return self._fig

    def new_file(self, filename: str) -> Path:
        """TBW."""
        new_filename = self.output_dir.joinpath(filename)  # type: Path
        new_filename.touch()

        return new_filename

    # TODO: Specific to 'single_plot' ?
    def save_to_png(
        self, data: np.ndarray, name: str, with_auto_suffix: bool = True
    ) -> Path:
        """Write array to bitmap PNG image file.

        Parameters
        ----------
        data : array
        name : str

        Returns
        -------
        Path
            TBW.
        """
        row, col = data.shape
        name = str(name).replace(".", "_")

        if with_auto_suffix:
            filename = apply_run_number(self.output_dir.joinpath(f"{name}_??.png"))
        else:
            filename = self.output_dir / f"{name}.png"

        full_filename = filename.resolve()  # type: Path

        # TODO: Create a new figure ! Check this
        dpi = 300
        self._fig.set_size_inches(min(col / dpi, 10.0), min(row / dpi, 10.0))

        # ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
        # ax.set_axis_off()
        #
        # fig.add_axes(ax)
        # plt.imshow(data, cmap="gray", extent=[0, col, 0, row])
        self._fig.savefig(full_filename, dpi=dpi)

        return full_filename

    def save_to_fits(
        self, data: np.ndarray, name: str, with_auto_suffix: bool = True
    ) -> Path:
        """Write array to FITS file."""
        name = str(name).replace(".", "_")

        if with_auto_suffix:
            filename = apply_run_number(self.output_dir.joinpath(f"{name}_??.fits"))
        else:
            filename = self.output_dir / f"{name}.fits"

        full_filename = filename.resolve()  # type: Path
        self._log.info("Save to FITS - filename: '%s'", full_filename)

        hdu = fits.PrimaryHDU(data)
        hdu.header["PYXEL_V"] = (str(version), "Pyxel version")
        hdu.writeto(full_filename, overwrite=False, output_verify="exception")

        return full_filename

    def save_to_hdf(
        self, data: "Detector", name: str, with_auto_suffix: bool = True
    ) -> Path:
        """Write detector object to HDF5 file."""
        name = str(name).replace(".", "_")

        if with_auto_suffix:
            filename = apply_run_number(self.output_dir.joinpath(f"{name}_??.h5"))
        else:
            filename = self.output_dir / f"{name}.h5"

        full_filename = filename.resolve()  # type: Path

        with h5.File(full_filename, "w") as h5file:
            h5file.attrs["pyxel-version"] = str(version)
            if name == "detector":
                detector_grp = h5file.create_group("detector")
                for array, name in zip(
                    [
                        data.signal.array,
                        data.image.array,
                        data.photon.array,
                        data.pixel.array,
                        data.charge.frame,
                    ],
                    ["Signal", "Image", "Photon", "Pixel", "Charge"],
                ):
                    dataset = detector_grp.create_dataset(name, np.shape(array))
                    dataset[:] = array
            else:
                detector_grp = h5file.create_group("data")
                dataset = detector_grp.create_dataset(name, np.shape(data))
                dataset[:] = data
        return filename

    def save_to_txt(
        self, data: np.ndarray, name: str, with_auto_suffix: bool = True
    ) -> Path:
        """Write data to txt file."""
        name = str(name).replace(".", "_")

        if with_auto_suffix:
            filename = apply_run_number(self.output_dir.joinpath(f"{name}_??.txt"))
        else:
            filename = self.output_dir / f"{name}.txt"

        full_filename = filename.resolve()  # type: Path
        np.savetxt(full_filename, data, delimiter=" | ", fmt="%.8e")

        return full_filename

    def save_to_csv(
        self, data: pd.DataFrame, name: str, with_auto_suffix: bool = True
    ) -> Path:
        """Write Pandas Dataframe or Numpy array to a CSV file."""
        name = str(name).replace(".", "_")

        if with_auto_suffix:
            filename = apply_run_number(self.output_dir.joinpath(f"{name}_??.csv"))
        else:
            filename = self.output_dir / f"{name}.csv"

        full_filename = filename.resolve()
        try:
            data.to_csv(full_filename, float_format="%g")
        except AttributeError:
            np.savetxt(full_filename, data, delimiter=",", fmt="%.8e")

        return full_filename

    def save_to_npy(
        self, data: np.ndarray, name: str, with_auto_suffix: bool = True
    ) -> Path:
        """Write Numpy array to Numpy binary npy file."""
        name = str(name).replace(".", "_")

        if with_auto_suffix:
            filename = apply_run_number(self.output_dir.joinpath(f"{name}_??.npy"))
        else:
            filename = self.output_dir / f"{name}.npy"

        full_filename = filename.resolve()  # type: Path

        np.save(file=full_filename, arr=data)
        return full_filename

    def save_plot(self, filename: str = "figure_??") -> Path:
        """Save plot figure in PNG format, close figure and create new canvas for next plot."""
        new_filename = self.output_dir.joinpath(filename + ".png")  # type: Path
        output_filename = apply_run_number(new_filename).resolve()  # type: Path

        self._log.info("Save plot in filename '%s'", output_filename)
        self._fig.savefig(output_filename)
        # plt.close('all')
        # plt.figure()

        return output_filename

    def plot_graph(
        self, x: np.ndarray, y: np.ndarray, args: t.Optional[dict] = None
    ) -> None:
        """TBW.

        Parameters
        ----------
        x
        y
        args
        """
        ax_args0, plt_args0 = self.update_args(plot_type=PlotType.Graph, new_args=args)

        user_plt_args_dct = None  # type: t.Optional[t.Mapping]
        if isinstance(self.user_plt_args, PlotArguments):
            user_plt_args_dct = self.user_plt_args.to_dict()

        ax_args, plt_args = self.update_args(
            plot_type=PlotType.Graph,
            new_args=user_plt_args_dct,
            ax_args=ax_args0,
            plt_args=plt_args0,
        )

        self._ax.plot(
            x,
            y,
            color=plt_args["color"],
            marker=plt_args["marker"],
            linestyle=plt_args["linestyle"],
        )
        update_plot(ax_args=ax_args, ax=self._ax)
        # plt.draw()

    def plot_histogram(self, data: np.ndarray, args: t.Optional[dict] = None) -> None:
        """TBW.

        Parameters
        ----------
        data
        args
        """
        assert self.user_plt_args is not None

        user_plt_args_dct = None  # type: t.Optional[t.Mapping]
        if isinstance(self.user_plt_args, PlotArguments):
            user_plt_args_dct = self.user_plt_args.to_dict()

        ax_args0, plt_args0 = self.update_args(
            plot_type=PlotType.Histogram, new_args=args
        )
        ax_args, plt_args = self.update_args(
            plot_type=PlotType.Histogram,
            new_args=user_plt_args_dct,
            ax_args=ax_args0,
            plt_args=plt_args0,
        )
        if isinstance(data, np.ndarray):
            data = data.flatten()

        self._ax.hist(
            x=data,
            bins=plt_args["bins"],
            range=plt_args["range"],
            density=plt_args["density"],
            log=plt_args["log"],
            cumulative=plt_args["cumulative"],
            histtype=plt_args["histtype"],
            color=plt_args["color"],
            facecolor=plt_args["facecolor"],
        )

        update_plot(ax_args=ax_args, ax=self._ax)

    def plot_scatter(
        self,
        x: np.ndarray,
        y: np.ndarray,
        color: t.Optional[str] = None,
        args: t.Optional[dict] = None,
    ) -> None:
        """TBW.

        Parameters
        ----------
        x
        y
        color
        args
        """
        user_plt_args_dct = None  # type: t.Optional[t.Mapping]

        if isinstance(self.user_plt_args, PlotArguments):
            user_plt_args_dct = self.user_plt_args.to_dict()

        ax_args0, plt_args0 = self.update_args(
            plot_type=PlotType.Scatter, new_args=args
        )
        ax_args, plt_args = self.update_args(
            plot_type=PlotType.Scatter,
            new_args=user_plt_args_dct,
            ax_args=ax_args0,
            plt_args=plt_args0,
        )

        # fig = plt.gcf()
        # ax = fig.axes

        if color is not None:
            sp = self._ax.scatter(x, y, c=color, s=plt_args["size"])
            cbar = self._fig.colorbar(sp)
            cbar.set_label(plt_args["cbar_label"])

        else:
            self._ax.scatter(x, y, s=plt_args["size"])

        update_plot(ax_args=ax_args, ax=self._ax)
        # plt.draw()
        # fig.canvas.draw_idle()

    def save_to_file(
        self,
        processor: "Processor",
        prefix: t.Optional[str] = None,
        with_auto_suffix: bool = True,
    ) -> t.Sequence[Path]:
        """Save outputs into file(s).

        Parameters
        ----------
        processor : Processor

        Returns
        -------
        list of ``Path``
            TBW.
        """
        save_methods = {
            "fits": self.save_to_fits,
            "hdf": self.save_to_hdf,
            "npy": self.save_to_npy,
            "txt": self.save_to_txt,
            "csv": self.save_to_csv,
            "png": self.save_to_png,
        }  # type: t.Mapping[ValidFormat, SaveToFile]

        filenames = []  # type: t.List[Path]

        dct: t.Mapping[ValidName, t.Sequence[ValidFormat]]
        for dct in self.save_data_to_file:
            # TODO: Why looking at first entry ? Check this !
            # Get first entry of `dict` 'item'
            first_item: t.Tuple[ValidName, t.Sequence[ValidFormat]]
            first_item, *_ = dct.items()

            obj: ValidName
            format_list: t.Sequence[ValidFormat]
            obj, format_list = first_item

            data = processor.get(obj)  # type: np.ndarray

            if prefix:
                name = f"{prefix}_{obj}"  # type: str
            else:
                name = obj

            out_format: ValidFormat
            for out_format in format_list:
                func = save_methods[out_format]  # type: SaveToFile
                filename = func(
                    data=data, name=name, with_auto_suffix=with_auto_suffix
                )  # type: Path

                filenames.append(filename)

        return filenames

    # TODO: Specific to 'calibration_plot'
    def champions_plot(
        self,
        results: t.Mapping[str, t.Union[Number, np.ndarray]],
        champions_file: Path,
        island_id: int,
    ) -> None:
        """TBW."""
        data = np.loadtxt(champions_file)
        generations = data[:, 0].astype(int)
        title = "Calibrated parameter: "

        a = 1
        for key, param_value in results.items():
            plt_args = {"xlabel": "generation", "linestyle": "-", "sci_y": True}

            param_name = key[slice(key.rfind(".") + 1, None)]
            plt_args["ylabel"] = param_name

            if param_name == "fitness":
                plt_args["title"] = "Champion fitness"
                plt_args["color"] = "red"
                plt_args["ylabel"] = "fitness"

            elif param_name == "island":
                continue

            else:
                if key.rfind(".arguments") == -1:
                    mdn = key[: key.rfind("." + param_name)]
                else:
                    mdn = key[: key.rfind(".arguments")]
                model_name = mdn[slice(mdn.rfind(".") + 1, None)]
                plt_args["title"] = title + model_name + " / " + param_name
                plt_args["ylabel"] = param_name

            b = 1
            if isinstance(param_value, Number):
                column = data[:, a]
                self.plot_graph(x=generations, y=column, args=plt_args)

            elif isinstance(param_value, np.ndarray):
                b = len(param_value)
                column = data[:, slice(a, a + b)]
                self.plot_graph(x=generations, y=column, args=plt_args)
                self._fig.legend([f"index {i}" for i in range(b)])

            self.save_plot(filename=f"calibrated_{param_name!s}_id{island_id!s}")
            a += b

    # TODO: Specific to 'calibration_plot' ??
    def population_plot(
        self,
        # results: dict,
        population_file: Path,
        island_id: int,
    ) -> None:
        """TBW."""
        assert self.calibration_plot

        data = np.loadtxt(population_file)  # type: np.ndarray
        fitnesses = np.log10(data[:, 1])
        a, b = 2, 1  # 1st parameter and fitness
        if self.calibration_plot.population_plot:
            if self.calibration_plot.population_plot.columns:
                col = self.calibration_plot.population_plot.columns
                a, b = col[0], col[1]
        x = data[:, a]
        y = data[:, b]

        plt_args = {
            "title": "Population of the last generation",
            "size": 8,
            "cbar_label": "log(fitness)",
        }
        if b == 0:
            plt_args["ylabel"] = "generation"
        elif b == 1:
            plt_args["ylabel"] = "fitness"
        else:
            plt_args["ylabel"] = "champions file column #" + str(b)

        if a == 0:
            plt_args["xlabel"] = "generation"
        elif a == 1:
            plt_args["xlabel"] = "fitness"
        else:
            plt_args["xlabel"] = "champions file column #" + str(a)

        if a == 1 or b == 1:
            plt_args["sci_y"] = True
            self.plot_scatter(x, y, args=plt_args)
        else:
            self.plot_scatter(x, y, color=fitnesses, args=plt_args)

        self.save_plot(filename=f"population_id{island_id}")

    # TODO: Specific to 'calibration_plot'
    def calibration_outputs(self, processor_list: "t.Sequence[Processor]") -> None:
        """TBW."""
        if self.save_data_to_file:
            for processor in processor_list:
                self.save_to_file(processor)

                # if self._single_plot:
                #    self.single_to_plot(processor)

    def save_processors(self, processors: pd.DataFrame) -> t.Sequence[Delayed]:
        """TBW."""
        lst = []  # type: t.List[delayed.Delayed]

        if self.save_data_to_file:

            for _, serie in processors.iterrows():
                id_island = serie["island"]  # type: int
                id_processor = serie["id_processor"]  # type: int
                processor = serie["processor"]  # type: Delayed

                # TODO: Create folders ?
                prefix = f"island{id_island:02d}_processor{id_processor:02d}"

                output_filenames = delayed(self.save_to_file)(
                    processor=processor, prefix=prefix, with_auto_suffix=False
                )  # type: Delayed
                lst.append(output_filenames)

        return lst

    # TODO: Specific to 'calibration_plot'
    # def calibration_plots(self, results: t.Mapping, fitness: float) -> None:
    #     """TBW."""
    #     assert self.calibration_plot
    #
    #     if self.calibration_plot:
    #         if self.calibration_plot.champions_plot:
    #             self.user_plt_args = None
    #
    #             if self.calibration_plot.champions_plot.plot_args:
    #                 self.user_plt_args = self.calibration_plot.champions_plot.plot_args
    #
    #             for iid, file_ch in enumerate(
    #                 self.output_dir.glob("champions_id*.out")
    #             ):
    #                 self.champions_plot(
    #                     results={"fitness": fitness, **results},
    #                     champions_file=file_ch,
    #                     island_id=iid,
    #                 )
    #
    #         if self.calibration_plot.population_plot:
    #             self.user_plt_args = None
    #             if self.calibration_plot.population_plot.plot_args:
    #                 self.user_plt_args = self.calibration_plot.population_plot.plot_args
    #
    #             for iid, file_pop in enumerate(
    #                 self.output_dir.glob("population_id*.out")
    #             ):
    #                 self.population_plot(
    #                     # results=results,
    #                     population_file=file_pop,
    #                     island_id=iid,
    #                 )

    # TODO: Specific to 'calibration_plot' ??
    def fitting_plot(
        self, target_data: np.ndarray, simulated_data: np.ndarray, data_i: int
    ) -> None:
        """TBW."""
        assert self.calibration_plot

        if self.calibration_plot.fitting_plot:
            self._ax.plot(target_data, ".-", label=f"target data #{data_i}")
            self._ax.plot(simulated_data, ".-", label=f"simulated data #{data_i}")
            self._fig.canvas.draw_idle()

    # TODO: Specific to 'calibration_plot' ??
    def fitting_plot_close(self, result_type: "ResultType", island: int) -> None:
        """TBW."""
        assert self.calibration_plot

        if self.calibration_plot.fitting_plot:
            self.user_plt_args = None
            if self.calibration_plot.fitting_plot.plot_args is None:
                raise RuntimeError

            self.user_plt_args = self.calibration_plot.fitting_plot.plot_args

            user_plt_args_dct = None  # type: t.Optional[t.Mapping[str, t.Any]]
            if isinstance(self.user_plt_args, PlotArguments):
                user_plt_args_dct = self.user_plt_args.to_dict()

            args = {
                "title": f"Target and Simulated ({result_type}) "
                f"data, "
                f"island {island}"
            }
            ax_args0, plt_args0 = self.update_args(
                plot_type=PlotType.Graph, new_args=args
            )
            ax_args, plt_args = self.update_args(
                plot_type=PlotType.Graph,
                new_args=user_plt_args_dct,
                ax_args=ax_args0,
                plt_args=plt_args0,
            )
            update_plot(ax_args=ax_args, ax=self._ax)
            # plt.legend()
            self._fig.legend()

            self.save_plot(filename=f"fitted_datasets_id{island}")

    def update_args(
        self,
        plot_type: PlotType,
        new_args: t.Optional[t.Mapping] = None,
        ax_args: t.Optional[dict] = None,
        plt_args: t.Optional[dict] = None,
    ) -> t.Tuple[dict, dict]:
        """TBW."""
        if new_args is None:
            new_args = {}

        if ax_args is None:
            ax_args = self.default_ax_args.to_dict()

        if plt_args is None:
            if plot_type is PlotType.Histogram:
                plt_args = self.default_hist_args.copy()
            elif plot_type is PlotType.Graph:
                plt_args = self.default_plot_args.copy()
            elif plot_type is PlotType.Scatter:
                plt_args = self.default_scatter_args.copy()
            else:
                raise ValueError

        for key in new_args:
            if key in plt_args.keys():
                plt_args[key] = new_args[key]
            elif key in ax_args.keys():
                ax_args[key] = new_args[key]
            else:
                raise KeyError('Not valid plotting key in "plot_args": "%s"' % key)

        return ax_args, plt_args



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Subpackage to define options in Pyxel."""
import typing as t
from pathlib import Path

import attr


@attr.s(
    auto_attribs=True,
    on_setattr=attr.setters.pipe(attr.setters.convert, attr.setters.validate),
)
class GlobalOptions:
    """Define a container class for all available options."""

    cache_enabled: bool = attr.ib(
        validator=attr.validators.instance_of(bool),
        default=False,
    )
    cache_folder: t.Optional[t.Union[str, Path]] = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of((str, Path))),
        default=None,
    )

    def update(self, dct: t.Mapping) -> t.Mapping:
        """Apply the option(s) in this container class.

        Parameters
        ----------
        dct : Mapping
            Validated option(s) to apply.

        Returns
        -------
        Mapping
            Previous option(s).

        Examples
        --------
        >>> options = GlobalOptions()
        >>> options
        GlobalOptions(cache_enabled=False, cache_folder=None)

        >>> options.update({"cache_enabled: True"})
        {'cache_enabled': False}
        >>> options
        GlobalOptions(cache_enabled=True, cache_folder=None)
        """
        previous_params = {}
        for key, value in dct.items():
            previous_params[key] = getattr(self, key)

            setattr(self, key, value)

        return previous_params

    def validate_and_convert(self, dct: t.Mapping) -> t.Mapping:
        """Validate and convert the input 'option(s)'.

        Parameters
        ----------
        dct : Mapping
            The input option(s) to validate and to convert.

        Returns
        -------
        Mapping
            The validated and converted option(s)

        Examples
        --------
        >>> options = GlobalOptions()
        >>> options
        GlobalOptions(cache_enabled=False, cache_folder=None)

        >>> options.validate_and_convert({"cache_enabled": True})
        {'cache_enabled': True}
        >>> options.validate_and_convert({"cache_enabled": "hello"})
        TypeError(...)
        """
        valid_keys = list(attr.asdict(self, recurse=False))

        result = {}
        for key, value in dct.items():
            if key not in valid_keys:
                raise KeyError(f"Wrong option {key!r}.")

            # TODO: Check and convert types
            result[key] = value

        return result


# Create a singleton
global_options = GlobalOptions()


class SetOptions:
    """Set options for Pyxel in a controlled context.

    Currently supported options:

    * ``cache_enabled``: Enable/disable caching when reading table and data.
      Default value: False
    * ``cache_folder``: Set a folder for caching files.
      Default value: None

    Examples
    --------
    Set global options
    >>> pyxel.set_options(cache_folder="/tmp")

    Or with a context manager
    >>> with pyxel.set_options(cache_enabled=True, cache_folder="/tmp"):
    ...     print("Do something")
    """

    def __init__(self, **kwargs):
        valid_parameters = global_options.validate_and_convert(
            kwargs
        )  # type: t.Mapping

        self._previous_params = global_options.update(
            valid_parameters
        )  # type: t.Mapping

    def __enter__(self) -> "SetOptions":
        return self

    def __exit__(self, exc_type: t.Any, exc_value: t.Any, traceback: t.Any) -> None:
        _ = global_options.update(self._previous_params)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Fitness functions for model fitting."""

import typing as t

import numpy as np


def sum_of_abs_residuals(
    simulated: np.ndarray, target: np.ndarray, weighting: t.Optional[float] = None
) -> np.ndarray:
    """TBW.

    Parameters
    ----------
    simulated
    target
    weighting

    Returns
    -------
    array
        TBW.
    """
    try:
        diff = target.astype(float) - simulated.astype(float)  # type: np.ndarray
    except AttributeError:
        diff = float(target) - float(simulated)

    if weighting is not None:
        diff *= weighting
    return np.sum(np.abs(diff))


def sum_of_squared_residuals(
    simulated: np.ndarray, target: np.ndarray, weighting: t.Optional[float] = None
) -> np.ndarray:
    """TBW.

    Parameters
    ----------
    simulated
    target
    weighting

    Returns
    -------
    array
        TBW.
    """
    try:
        diff = target.astype(float) - simulated.astype(float)
    except AttributeError:
        diff = float(target) - float(simulated)
    diff_square = diff * diff

    if weighting is not None:
        diff_square *= weighting
    return np.sum(diff_square)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Subpackage to load images and tables."""

import typing as t
from pathlib import Path

import fsspec
import numpy as np
import pandas as pd
from astropy.io import fits
from PIL import Image

from pyxel.options import global_options


def load_image(filename: t.Union[str, Path]) -> np.ndarray:
    """Load a 2D image.

    Parameters
    ----------
    filename : str or Path
        Filename to read an image.
        {.npy, .fits, .txt, .data, .jpg, .jpeg, .bmp, .png, .tiff} are accepted.

    Returns
    -------
    array : ndarray
        A 2D array.

    Raises
    ------
    FileNotFoundError
        If an image is not found.
    ValueError
        When the extension of the filename is unknown or separator is not found.

    Examples
    --------
    >>> from pyxel.inputs_outputs import load_image
    >>> load_image("frame.fits")
    array([[-0.66328494, -0.63205819, ...]])

    >>> load_image("https://hostname/folder/frame.fits")
    array([[-0.66328494, -0.63205819, ...]])

    >>> load_image("another_frame.npy")
    array([[-1.10136521, -0.93890239, ...]])

    >>> load_image("rgb_frame.jpg")
    array([[234, 211, ...]])
    """
    # Extract suffix (e.g. '.txt', '.fits'...)
    suffix = Path(filename).suffix.lower()  # type: str

    if isinstance(filename, Path):
        full_filename = filename.expanduser().resolve()  # type: Path
        if not full_filename.exists():
            raise FileNotFoundError(f"Input file '{full_filename}' can not be found.")

        url_path = str(full_filename)  # type: str

    else:
        url_path = filename

    # Define extra parameters to use with 'fsspec'
    extras = {}
    if global_options.cache_enabled:
        url_path = f"simplecache::{url_path}"

        if global_options.cache_folder:
            extras["simplecache"] = {"cache_storage": global_options.cache_folder}

    if suffix.startswith(".fits"):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            data_2d = fits.getdata(file_handler)  # type: np.ndarray

    elif suffix.startswith(".npy"):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            data_2d = np.load(file_handler)

    elif suffix.startswith(".txt") or suffix.startswith(".data"):
        for sep in ["\t", " ", ",", "|", ";"]:
            try:
                with fsspec.open(url_path, mode="r", **extras) as file_handler:
                    data_2d = np.loadtxt(file_handler, delimiter=sep)
                break
            except ValueError:
                pass
        else:
            raise ValueError(f"Cannot find the separator for filename '{url_path}'.")

    elif suffix.startswith((".jpg", ".jpeg", ".png", ".bmp", ".tiff", ".tif")):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            image_2d = Image.open(file_handler)
            image_2d_converted = image_2d.convert("LA")  # RGB to grayscale conversion

        data_2d = np.array(image_2d_converted)[:, :, 0]

    else:
        raise ValueError(
            "Image format not supported. List of supported image formats: "
            ".npy, .fits, .txt, .data, .jpg, .jpeg, .bmp, .png, .tiff, .tif."
        )

    return data_2d


def load_table(filename: t.Union[str, Path]) -> pd.DataFrame:
    """Load a table from a file and returns a pandas dataframe. No header is expected in xlsx.

    Parameters
    ----------
    filename: str or Path
        Filename to read the table.
        {.npy, .xlsx, .csv, .txt., .data} are accepted.

    Returns
    -------
    table: DataFrame

    Raises
    ------
    FileNotFoundError
        If an image is not found.
    ValueError
        When the extension of the filename is unknown or separator is not found.

    """
    suffix = Path(filename).suffix.lower()  # type: str

    if isinstance(filename, Path):
        full_filename = Path(filename).expanduser().resolve()
        if not full_filename.exists():
            raise FileNotFoundError(f"Input file '{full_filename}' can not be found.")

        url_path = str(full_filename)  # type: str
    else:
        url_path = filename

    # Define extra parameters to use with 'fsspec'
    extras = {}
    if global_options.cache_enabled:
        url_path = f"simplecache::{url_path}"

        if global_options.cache_folder:
            extras["simplecache"] = {"cache_storage": global_options.cache_folder}

    if suffix.startswith(".npy"):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            table = pd.DataFrame(np.load(file_handler), dtype="float")

    elif suffix.startswith(".xlsx"):
        with fsspec.open(url_path, mode="rb", **extras) as file_handler:
            table = pd.read_excel(
                file_handler, header=None, convert_float=False, engine="openpyxl"
            )

    elif suffix.startswith(".csv"):
        for sep in ["\t", " ", ",", "|", ";"]:
            try:
                # numpy will return ValueError with a wrong delimiter
                with fsspec.open(url_path, mode="r", **extras) as file_handler:
                    table = pd.read_csv(
                        file_handler, delimiter=sep, header=None, dtype="float"
                    )
                    break
            except ValueError:
                pass
        else:
            raise ValueError("Cannot find the separator.")

    elif suffix.startswith(".txt") or suffix.startswith(".data"):
        for sep in ["\t", " ", ",", "|", ";"]:
            try:
                with fsspec.open(url_path, mode="r", **extras) as file_handler:
                    table = pd.read_table(
                        file_handler, delimiter=sep, header=None, dtype="float"
                    )
                    break
            except ValueError:
                pass
        else:
            raise ValueError("Cannot find the separator.")

    else:
        raise ValueError("Only .npy, .xlsx, .csv, .txt and .data implemented.")

    return table



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import itertools
import logging
import typing as t
from copy import deepcopy
from enum import Enum

import numpy as np

from pyxel.parametric.parameter_values import ParameterValues
from pyxel.state import get_obj_att, get_value

if t.TYPE_CHECKING:
    from ..inputs_outputs import ParametricOutputs
    from ..pipelines import Processor


class ParametricMode(Enum):
    """TBW."""

    Embedded = "embedded"
    Sequential = "sequential"
    Parallel = "parallel"


# TODO: Use `Enum` for `parametric_mode` ?
class Parametric:
    """TBW."""

    def __init__(
        self,
        outputs: "ParametricOutputs",
        mode: str,
        parameters: t.Sequence[ParameterValues],
        from_file: t.Optional[str] = None,
        column_range: t.Optional[t.Tuple[int, int]] = None,
    ):
        self.outputs = outputs
        self.parametric_mode = ParametricMode(mode)  # type: ParametricMode
        self._parameters = parameters
        self.file = from_file
        self.data = None  # type: t.Optional[np.ndarray]
        if column_range:
            self.columns = slice(*column_range)

    def __repr__(self):
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<mode={self.parametric_mode!s}>"

    @property
    def enabled_steps(self) -> t.Sequence[ParameterValues]:
        """TBW."""
        return [step for step in self._parameters if step.enabled]

    def _parallel(self, processor: "Processor") -> "t.Iterator[Processor]":
        """TBW.

        :param processor:
        :return:
        """
        self.data = np.loadtxt(self.file)[:, self.columns]
        for data_array in self.data:
            i = 0
            new_proc = deepcopy(processor)  # type: Processor
            for step in self.enabled_steps:
                key = step.key

                # TODO: this is confusing code. Fix this.
                #       Furthermore 'step.values' should be a `t.List[int, float]` and not a `str`
                if step.values == "_":
                    value = data_array[i]
                    i += 1

                elif isinstance(step.values, list) and all(
                    x == "_" for x in step.values[:]
                ):
                    value = data_array[i : i + len(step.values)]  # noqa: E203
                    i += len(value)

                else:
                    raise ValueError(
                        'Only "_" characters (or a list of them) should be used to '
                        "indicate parameters updated from file in parallel"
                    )

                new_proc.set(key, value)
            yield new_proc

    def _sequential(self, processor: "Processor") -> "t.Iterator[Processor]":
        """TBW.

        :param processor:
        :return:
        """
        for step in self.enabled_steps:  # type: ParameterValues
            key = step.key  # type : str
            for value in step:
                # step.current = value
                new_proc = deepcopy(processor)  # type: Processor
                new_proc.set(key=key, value=value)
                yield new_proc

    def _embedded(self, processor: "Processor") -> "t.Iterator[Processor]":
        """TBW.

        :param processor:
        :return:
        """
        all_steps = self.enabled_steps
        keys = [step.key for step in self.enabled_steps]
        for params in itertools.product(*all_steps):
            new_proc = deepcopy(processor)  # type: Processor
            for key, value in zip(keys, params):
                # for step in all_steps:
                #     if step.key == key:
                #         step.current = value
                new_proc.set(key=key, value=value)
            yield new_proc

    def collect(self, processor: "Processor") -> "t.Iterator[Processor]":
        """TBW."""
        for step in self.enabled_steps:  # type: ParameterValues

            # TODO: the string literal expressions are difficult to maintain.
            #     Example: 'pipeline.', '.arguments', '.enabled'
            #     We may want to consider an API for this.
            # Proposed API:
            # value = operator.attrgetter(step.key)(processor)
            if "pipeline." in step.key:
                model_name = step.key[: step.key.find(".arguments")]  # type: str
                model_enabled = model_name + ".enabled"  # type: str
                if not processor.get(model_enabled):
                    raise ValueError(
                        f"The '{model_name}' model referenced in parametric configuration "
                        f"has not been enabled in yaml config!"
                    )

            if (
                any(x == "_" for x in step.values[:])
                and self.parametric_mode != ParametricMode.Parallel
            ):
                raise ValueError(
                    "Either define 'parallel' as parametric mode or "
                    "do not use '_' character in 'values' field"
                )

        if self.parametric_mode == ParametricMode.Embedded:
            configs_it = self._embedded(processor)  # type: t.Iterator[Processor]

        elif self.parametric_mode == ParametricMode.Sequential:
            configs_it = self._sequential(processor)

        elif self.parametric_mode == ParametricMode.Parallel:
            configs_it = self._parallel(processor)

        else:
            # configs_it = iter([])
            raise NotImplementedError()

        return configs_it

    def debug(self, processor: "Processor") -> list:
        """TBW."""
        result = []
        configs = self.collect(processor)
        for i, config in enumerate(configs):
            values = []
            for step in self.enabled_steps:
                _, att = get_obj_att(config, step.key)
                value = get_value(config, step.key)
                values.append((att, value))
            logging.debug("%d: %r" % (i, values))
            result.append((i, values))
        return result


# # TODO: Use a `Enum` for 'mode' ?
# # TODO: Create several classes `ConfigurationSingle`, `ConfigurationParametric`,
# #       `ConfigurationCalibration` and `ConfigurationDynamic`
# # TODO: Move this class into its own file 'configuration.py'
# class Configuration:
#     """TBW."""
#
#     def __init__(
#         self,
#         mode: str,
#         outputs: "Outputs",
#         parametric: t.Optional[Parametric] = None,
#         calibration: "t.Optional[Calibration]" = None,
#         dynamic: t.Optional[t.Dict[str, t.Any]] = None,
#     ):
#         if mode not in ["single", "parametric", "calibration", "dynamic"]:
#             raise ValueError(
#                 "Non-existing running mode defined for Pyxel in yaml config file."
#             )
#
#         self.mode = mode  # type: str
#
#         self.outputs = outputs  # type: Outputs
#
#         self.parametric = parametric  # type: t.Optional[Parametric]
#         self.calibration = calibration  # type: t.Optional[Calibration]
#         self.dynamic = dynamic  # type: t.Optional[t.Dict[str, t.Any]]
#
#         if mode == "parametric":
#             assert self.parametric
#             self.outputs.params_func(self.parametric)
#
#     def __repr__(self) -> str:
#         cls_name = self.__class__.__name__  # type: str
#
#         return f"{cls_name}<mode={self.mode!r}, outputs={self.outputs!r}>"



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel charge injection functions for CCDs."""
import logging
import typing as t

import numpy as np

from pyxel.detectors import CCD


# TODO: Fix this
# @validators.validate
# @config.argument(name='detector', label='', units='', validate=checkers.check_type(CCD))
def charge_blocks(
    detector: CCD,
    charge_level: int,
    block_start: int = 0,
    block_end: t.Optional[int] = None,
) -> None:
    """TBW.

    :param detector:
    :param charge_level:
    :param block_start:
    :param block_end:
    :return:
    """
    logging.info("")
    geo = detector.geometry
    if block_end is None:
        block_end = geo.row

    # all pixels has zero charge by default
    detector_charge = np.zeros((geo.row, geo.col))
    detector_charge[slice(block_start, block_end), :] = charge_level
    charge_number = detector_charge.flatten()
    where_non_zero = np.where(charge_number > 0.0)
    charge_number = charge_number[where_non_zero]
    size = charge_number.size

    init_ver_pix_position = geo.vertical_pixel_center_pos_list()[where_non_zero]
    init_hor_pix_position = geo.horizontal_pixel_center_pos_list()[where_non_zero]

    detector.charge.add_charge(
        particle_type="e",
        particles_per_cluster=charge_number,
        init_energy=[0.0] * size,
        init_ver_position=init_ver_pix_position,
        init_hor_position=init_hor_pix_position,
        init_z_position=[0.0] * size,
        init_ver_velocity=[0.0] * size,
        init_hor_velocity=[0.0] * size,
        init_z_velocity=[0.0] * size,
    )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import logging
import typing as t
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
from dask.delayed import Delayed
from typing_extensions import Literal

from pyxel.calibration import (
    Algorithm,
    CalibrationMode,
    DaskBFE,
    DaskIsland,
    Island,
    MyArchipelago,
    ResultType,
)
from pyxel.calibration.fitting import ModelFitting
from pyxel.parametric.parameter_values import ParameterValues
from pyxel.pipelines import ModelFunction, Processor

try:
    import pygmo as pg
except ImportError:
    import warnings

    warnings.warn("Cannot import 'pygmo", RuntimeWarning, stacklevel=2)

if t.TYPE_CHECKING:
    from ..inputs_outputs import CalibrationOutputs


def to_path_list(values: t.Sequence[t.Union[str, Path]]) -> t.List[Path]:
    """TBW."""
    return [Path(obj).resolve() for obj in values]


class Calibration:
    """TBW."""

    def __init__(
        self,
        outputs: "CalibrationOutputs",
        output_dir: t.Optional[Path] = None,
        fitting: t.Optional[ModelFitting] = None,
        mode: Literal["pipeline", "single_model"] = "pipeline",
        result_type: Literal["image", "signal", "pixel"] = "image",
        result_fit_range: t.Optional[t.Sequence[int]] = None,
        result_input_arguments: t.Optional[t.Sequence[ParameterValues]] = None,
        target_data_path: t.Optional[t.Sequence[Path]] = None,
        target_fit_range: t.Optional[t.Sequence[int]] = None,
        fitness_function: t.Optional[ModelFunction] = None,
        algorithm: t.Optional[Algorithm] = None,
        parameters: t.Optional[t.Sequence[ParameterValues]] = None,
        seed: t.Optional[int] = None,
        num_islands: int = 1,
        num_evolutions: int = 1,
        num_best_decisions: t.Optional[int] = None,
        topology: Literal["unconnected", "ring", "fully_connected"] = "unconnected",
        type_islands: Literal[
            "multiprocessing", "multithreading", "ipyparallel"
        ] = "multiprocessing",
        weighting_path: t.Optional[t.Sequence[Path]] = None,
    ):
        if seed is not None and seed not in range(100001):
            raise ValueError("'seed' must be between 0 and 100000.")

        if num_islands < 1:
            raise ValueError("'num_islands' must superior or equal to 1.")

        self._log = logging.getLogger(__name__)

        self.outputs = outputs

        self._output_dir = output_dir  # type:t.Optional[Path]
        self._fitting = fitting  # type: t.Optional[ModelFitting]

        self._calibration_mode = CalibrationMode(mode)

        self._result_type = ResultType(result_type)  # type: ResultType

        self._result_fit_range = (
            result_fit_range if result_fit_range else []
        )  # type: t.Sequence[int]

        self._result_input_arguments = (
            result_input_arguments if result_input_arguments else []
        )  # type: t.Sequence[ParameterValues]

        self._target_data_path = (
            to_path_list(target_data_path) if target_data_path else []
        )  # type: t.Sequence[Path]
        self._target_fit_range = (
            target_fit_range if target_fit_range else []
        )  # type: t.Sequence[int]

        self._fitness_function = fitness_function  # type: t.Optional[ModelFunction]
        self._algorithm = algorithm  # type: t.Optional[Algorithm]

        self._parameters = (
            parameters if parameters else []
        )  # type: t.Sequence[ParameterValues]

        self._seed = np.random.randint(0, 100000) if seed is None else seed  # type: int

        self._num_islands = num_islands  # type: int
        self._num_evolutions = num_evolutions  # type: int
        self._num_best_decisions = num_best_decisions  # type: t.Optional[int]
        self._type_islands = Island(type_islands)  # type:Island
        self._topology = (
            topology
        )  # type: t.Literal['unconnected', 'ring', 'fully_connected']

        self._weighting_path = weighting_path  # type: t.Optional[t.Sequence[Path]]

    @property
    def output_dir(self) -> Path:
        """TBW."""
        if not self._output_dir:
            raise RuntimeError("No 'output_dir' defined !")

        return self._output_dir

    @output_dir.setter
    def output_dir(self, value: Path) -> None:
        """TBW."""
        self._output_dir = value

    @property
    def fitting(self) -> ModelFitting:
        """TBW."""
        if not self._fitting:
            raise RuntimeError("No 'fitting' defined !")

        return self._fitting

    @fitting.setter
    def fitting(self, value: ModelFitting) -> None:
        """TBW."""
        self._fitting = value

    @property
    def calibration_mode(self) -> CalibrationMode:
        """TBW."""
        return self._calibration_mode

    @calibration_mode.setter
    def calibration_mode(self, value: CalibrationMode) -> None:
        """TBW."""
        self._calibration_mode = value

    @property
    def result_type(self) -> ResultType:
        """TBW."""
        return self._result_type

    @result_type.setter
    def result_type(self, value: ResultType) -> None:
        """TBW."""
        self._result_type = value

    @property
    def result_fit_range(self) -> t.Sequence[int]:
        """TBW."""
        return self._result_fit_range

    @result_fit_range.setter
    def result_fit_range(self, value: t.Sequence[int]) -> None:
        """TBW."""
        self._result_fit_range = value

    @property
    def result_input_arguments(self) -> t.Sequence[ParameterValues]:
        """TBW."""
        return self._result_input_arguments

    @result_input_arguments.setter
    def result_input_arguments(self, value: t.Sequence[ParameterValues]) -> None:
        """TBW."""
        self._result_input_arguments = value

    @property
    def target_data_path(self) -> t.Sequence[Path]:
        """TBW."""
        return self._target_data_path

    @target_data_path.setter
    def target_data_path(self, value: t.Sequence[Path]) -> None:
        """TBW."""
        self._target_data_path = value

    @property
    def target_fit_range(self) -> t.Sequence[int]:
        """TBW."""
        return self._target_fit_range

    @target_fit_range.setter
    def target_fit_range(self, value: t.Sequence[int]) -> None:
        """TBW."""
        self._target_fit_range = value

    @property
    def fitness_function(self) -> ModelFunction:
        """TBW."""
        if not self._fitness_function:
            raise RuntimeError("No 'fitness_function' defined !")

        return self._fitness_function

    @fitness_function.setter
    def fitness_function(self, value: ModelFunction) -> None:
        """TBW."""
        self._fitness_function = value

    @property
    def algorithm(self) -> Algorithm:
        """TBW."""
        if not self._algorithm:
            raise RuntimeError("No 'algorithm' defined !")

        return self._algorithm

    @algorithm.setter
    def algorithm(self, value: Algorithm) -> None:
        """TBW."""
        self._algorithm = value

    @property
    def parameters(self) -> t.Sequence[ParameterValues]:
        """TBW."""
        return self._parameters

    @parameters.setter
    def parameters(self, value: t.Sequence[ParameterValues]) -> None:
        """TBW."""
        self._parameters = value

    @property
    def seed(self) -> int:
        """TBW."""
        return self._seed

    @seed.setter
    def seed(self, value: int) -> None:
        """TBW."""
        if value not in range(100001):
            raise ValueError("'seed' must be between 0 and 100000.")

        self._seed = value

    @property
    def num_islands(self) -> int:
        """TBW."""
        return self._num_islands

    @num_islands.setter
    def num_islands(self, value: int) -> None:
        """TBW."""
        if value < 1:
            raise ValueError("'num_islands' must superior or equal to 1.")

        self._num_islands = value

    @property
    def num_evolutions(self) -> int:
        """TBW."""
        return self._num_evolutions

    @num_evolutions.setter
    def num_evolutions(self, value: int) -> None:
        """TBW."""
        self._num_evolutions = value

    @property
    def num_best_decisions(self) -> t.Optional[int]:
        """TBW."""
        return self._num_best_decisions

    @num_best_decisions.setter
    def num_best_decisions(self, value: t.Optional[int]) -> None:
        """TBW."""
        if isinstance(value, int) and value < 0:
            raise ValueError(
                "'num_best_decisions' must be 'None' or a positive integer"
            )

        self._num_best_decisions = value

    @property
    def topology(self) -> Literal["unconnected", "ring", "fully_connected"]:
        """TBW."""
        return self._topology

    @topology.setter
    def topology(self, value: t.Any) -> None:
        if value not in ["unconnected", "ring", "fully_connected"]:
            raise ValueError(
                "Expecting value: 'unconnected', 'ring' or 'fully_connected'"
            )

        self._topology = value

    @property
    def weighting_path(self) -> t.Optional[t.Sequence[Path]]:
        """TBW."""
        return self._weighting_path

    @weighting_path.setter
    def weighting_path(self, value: t.Sequence[Path]) -> None:
        """TBW."""
        self._weighting_path = value

    def run_calibration(
        self,
        processor: Processor,
        output_dir: Path,
        with_progress_bar: bool = True,
    ) -> t.Tuple[xr.Dataset, pd.DataFrame, pd.DataFrame]:
        """Run calibration pipeline."""
        pg.set_global_rng_seed(seed=self.seed)
        self._log.info("Seed: %d", self.seed)

        self.output_dir = output_dir

        self.fitting = ModelFitting(processor=processor, variables=self.parameters)
        self.fitting.configure(
            calibration_mode=self.calibration_mode,
            generations=self.algorithm.generations,
            population_size=self.algorithm.population_size,
            simulation_output=self.result_type,
            fitness_func=self.fitness_function,
            target_output=self.target_data_path,
            target_fit_range=self.target_fit_range,
            out_fit_range=self.result_fit_range,
            input_arguments=self.result_input_arguments,
            weighting=self.weighting_path,
            file_path=output_dir,
        )

        if self.num_islands > 1:  # default
            # Create an archipelago
            user_defined_island = DaskIsland()
            user_defined_bfe = DaskBFE()

            if self.topology == "unconnected":
                topo = pg.unconnected()
            elif self.topology == "ring":
                topo = pg.ring()
            elif self.topology == "fully_connected":
                topo = pg.fully_connected()
            else:
                raise NotImplementedError(f"topology {self.topology!r}")

            # Create a new archipelago
            # This operation takes some time ...
            my_archipelago = MyArchipelago(
                num_islands=self.num_islands,
                udi=user_defined_island,
                algorithm=self.algorithm,
                problem=self.fitting,
                pop_size=self.algorithm.population_size,
                bfe=user_defined_bfe,
                topology=topo,
                seed=self.seed,
                with_bar=with_progress_bar,
            )

            # Run several evolutions in the archipelago
            ds, df_processors, df_all_logs = my_archipelago.run_evolve(
                num_evolutions=self._num_evolutions,
                num_best_decisions=self._num_best_decisions,
            )

        else:
            raise NotImplementedError("Not implemented for 1 island.")

        self._log.info("Calibration ended.")
        return ds, df_processors, df_all_logs

    # TODO: This function must be improved
    # TODO: Speed-up this function
    # def old_post_processing(
    #     self,
    #     calib_results: t.Sequence[CalibrationResult],
    #     output: "CalibrationOutputs",
    # ) -> None:
    #     """TBW."""
    #     for one_calib_result in tqdm(calib_results):  # type: CalibrationResult
    #
    #         # TODO: Create a new method in output called '.save_processor(processor)'
    #         output.calibration_outputs(processor_list=one_calib_result.processors)
    #
    #         for idx, (processor, target_data) in enumerate(
    #             zip(one_calib_result.processors, self.fitting.all_target_data)
    #         ):
    #             simulated_data = self.fitting.get_simulated_data(
    #                 processor
    #             )  # type: np.ndarray
    #             output.fitting_plot(
    #                 target_data=target_data, simulated_data=simulated_data, data_i=idx
    #             )
    #
    #         output.fitting_plot_close(
    #             result_type=self.result_type, island=one_calib_result.island
    #         )
    #
    #     first_calib_result = calib_results[0]  # type: CalibrationResult
    #     output.calibration_plots(
    #         results=first_calib_result.results, fitness=first_calib_result.fitness
    #     )

    def post_processing(
        self,
        ds: xr.Dataset,
        df_processors: pd.DataFrame,
        output: "CalibrationOutputs",
    ) -> t.Sequence[Delayed]:
        """TBW."""
        filenames = output.save_processors(
            processors=df_processors
        )  # type: t.Sequence[Delayed]

        # TODO: Use output.fitting_plot ?
        # TODO: Use output.fitting_plot_close ?
        # TODO: Use output.calibration_plots ?

        return filenames



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Fix pattern noise model."""
import logging
import typing as t
from pathlib import Path

import numpy as np

from pyxel.detectors import Detector, Geometry


# TODO: Fix this
# @validators.validate
# @config.argument(name='', label='', units='', validate=)
def fix_pattern_noise(
    detector: Detector, pixel_non_uniformity: t.Union[str, Path, None] = None
) -> None:
    """Add fix pattern noise caused by pixel non-uniformity during charge collection.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    pixel_non_uniformity : str or Path
        path to an ascii file with and array.
    """
    geo = detector.geometry  # type: Geometry

    if pixel_non_uniformity is not None:
        if ".npy" in str(pixel_non_uniformity):
            pnu = np.load(pixel_non_uniformity)
        else:
            pnu = np.loadtxt(pixel_non_uniformity)
    else:
        filename = Path("data/pixel_non_uniformity.npy").resolve()

        if filename.exists():
            logging.warning(
                "'pixel_non_uniformity' file is not defined, "
                "using array from file: %s",
                filename,
            )
            pnu = np.load(filename)
        else:
            logging.warning(
                "'pixel_non_uniformity' file is not defined, "
                "generated random array to file: %s",
                filename,
            )

            # pnu = 0.99 + np.random.random((geo.row, geo.col)) * 0.02
            pnu = np.random.normal(loc=1.0, scale=0.03, size=(geo.row, geo.col))
            np.save(filename, pnu)

    pnu = pnu.reshape((geo.row, geo.col))

    detector.pixel.array = detector.pixel.array.astype(np.float64) * pnu  # TODO: dtype!



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""TBW."""

from .dynamic import Dynamic

__all__ = ["Dynamic"]



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#

# Before _sequential

from dask import delayed


def create_new_processor(
    processor_input: Processor, keys: t.Sequence, params: t.Mapping
) -> Processor:
    processor_output = deepcopy(processor_input)
    # new_key = deepcopy(keys)

    for key, value in zip(keys, params):
        processor_output.set(key=key, value=value)

    return processor_output


def collect(self, processor, with_dask: bool):

    delayed_processor = delayed(processor)

    lst = []

    all_steps = self.enabled_steps
    keys = (step.key for step in self.enabled_steps)  # tuple
    delayed_keys = delayed(keys)

    if with_dask:
        func = delayed(my_new_func)
    else:
        func = my_new_func

    # _embedded
    for params in itertools.product(*all_steps):
        delayed_params = delayed(params, pure=True)

        new_proc = func(
            processor_input=delayed_processor, keys=delayed_keys, params=delayed_params
        )

        lst.append(new_proc)

    return lst
